from typing import List, Dict
import logging

from ..filters.abstract_filter import AbstractFilter, SourceFilter
from .. import END_SOURCE_PIPELINE, KO_FILTER_TAG

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


class AbstractPipeline(object):

    def __init__(self) -> None:
        self._filters: List = []
        self._source: SourceFilter = None
        self._sink: AbstractFilter = None
        self._pipeline_status: bool = False

    def stop_pipeline(self) -> None :
        self._pipeline_status = False 

    def add_source(self, source: AbstractFilter) -> bool:
        if source is not None:
            self._source = source
            self._pipeline_status = True
            return True
        else:
            return False

    def add_sink(self, sink: AbstractFilter) -> bool:
        if sink is not None:
            self._sink = sink
            return True
        else:
            return False

    def add_filter(self, filter: AbstractFilter) -> bool:
        if filter is not None:
            self._filters.append(filter)
            return True
        else:
            return False

    def _process_sink(self, message: Dict) -> None:
        if self._sink is not None:
            if END_SOURCE_PIPELINE in message:
                self._sink.last_process(message)
            elif not KO_FILTER_TAG in message:
                self._sink.process(message)


    def execute(self) -> Dict:
        raise NotImplementedError('Needs to be implemented')

    def generate_last_message(self, message: Dict = {}) -> Dict:
        raise NotImplementedError('Needs to be implemented')


from abc import ABC, abstractmethod
from typing import Dict

from .. import END_SOURCE_PIPELINE


class AbstractFilter(ABC):

    def __init__(self, conf: Dict = {}):
        self.wrong_ob_cnt: int = 0

    @abstractmethod
    def process(self, message: Dict) -> Dict:
        raise NotImplementedError("process function must be implemented")

    def last_process(self, message: Dict) -> Dict:
        """
            This method implements a last action to do. Can be modify if needed.
        """
        return message


class SourceFilter(AbstractFilter):

    def __init__(self):
        self.gen = self.generator()

    def generator(self) -> Dict:
        pass

    def process(self) -> Dict:
        message = {}
        while True:
            try:
                return next(self.gen)
            except StopIteration:
                return {END_SOURCE_PIPELINE: 'True'}



from typing import Dict
import collections.abc
import yaml


def yaml_to_dict(file_path: str) -> Dict:
    with open(file_path, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
        return data
    return None


def update_nested_dict(d: Dict, u: Dict) -> Dict:
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_nested_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d



from typing import List, Dict
import logging

from .abstractpipeline import AbstractPipeline
from .. import END_SOURCE_PIPELINE, KO_FILTER_TAG

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


class Pipeline(AbstractPipeline):

    def __init__(self) -> None:
        super().__init__()


    def execute(self) -> Dict:
        while self._pipeline_status:
            message = self._source.process()
            logger.debug(f'message: {message}')
            if END_SOURCE_PIPELINE in message:  # No more data in the source
                self.stop_pipeline()
                break
            else:
                for filter in self._filters:
                    message = filter.process(message)
                    if KO_FILTER_TAG in message:
                        break
                if self._pipeline_status :
                    self._process_sink(message)

        return self.generate_last_message(message)

    def generate_last_message(self, message: Dict = {}) -> Dict:
        for filter in self._filters:
            message = filter.last_process(message)
            if KO_FILTER_TAG in message:
                break
        self._process_sink(message)
        logger.debug(f'message : {message}')
        return message



from typing import Dict
import logging
import copy

from ...pipe import SUB_PIPELINE
from ..abstract_filter import AbstractFilter
from ...factory import pipeline_factory
from ...utils.to_dict import update_nested_dict
from ... import END_SOURCE_PIPELINE, GENERAL_CONF_PIPELINE


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


class SubPipelineFilter(AbstractFilter):
    """
        A filter that exectutes a pipeline each time self.process() is called.
    """

    def __init__(self, conf: Dict = {}):
        super().__init__()

        self.conf_filter = conf.get('conf_filter')
        self.catalog_filter = conf.get('catalog_filter')
        self.source_name = self.conf_filter.get(GENERAL_CONF_PIPELINE).get('source')

    def update_conf_pipeline(self, message) -> object:
        raise NotImplementedError

    def _update_conf_pipeline(self, message) -> object:
        new_conf = copy.deepcopy(self.conf_filter)
        # Modify the source input based on the message content
        message_conf = self.update_conf_pipeline(message)
        return update_nested_dict(new_conf, message_conf)

    def process(self, message: object) -> object:
        logger.debug(f'-- {self.__class__.__name__}: new message received (message = {message})')
        # Create a new pipeline and update the source

        curr_pipeline_conf: Dict = self._update_conf_pipeline(message)  # Update some configuration pipeline
        pipeline = pipeline_factory(conf=curr_pipeline_conf, catalog_filter=self.catalog_filter, pipeline_type = SUB_PIPELINE)

        # Launch the pipeline
        pipeline_message, stop_message = pipeline.execute()
        message.update(stop_message)
        message.update(pipeline_message)
        logger.debug(f'-- {self.__class__.__name__}: processed message = {message} {pipeline_message}')

        del curr_pipeline_conf
        del pipeline


        return message

        




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['filtering_pipeline',
 'filtering_pipeline.filters',
 'filtering_pipeline.filters.catalog_filter',
 'filtering_pipeline.filters.catalog_source',
 'filtering_pipeline.pipe',
 'filtering_pipeline.utils']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml==6.0']

setup_kwargs = {
    'name': 'filtering-pipeline',
    'version': '0.2.2',
    'description': 'This package implements the design pattern tube and filters for making AI pipeline pre- and post-processing',
    'long_description': '# Preprocessing\n\n\n\n## Installation\n\n\nWe use conda as an environment manager and poetry as dependency manager.\n\n1. Generate a conda env \nFirst, create and activate a basic conda env from the [env_prep.yml](./env/env_prep.yml) file. \n\nRun \n```\n    conda env create -f ./env/env_prep.yml\n```\n\nthen \n\n```\n    conda activate env_prep\n```\n\nNB: it can be good to change the conda name env into [env_basic_conda.yml](./env/env_basic_conda.yml) file.\n\n\n2. Install poetry and package dependencies\n\nTo install package dependencies with poetry, \n\n```\n    poetry install\n```\n\nTo update package dependencies, \n```\n    poetry update\n```\n\n\n## Testing \n\nFor running all the tests:\n\n```\n    poetry run pytest \n```\n\nFor running a specific test: \n```\n    poetry run pytest path/my_test\n```\n\n\nSee test coverage : [TO COMPLETE]\n\n\n## Preprocessing pipeline \n\n- [How does the preprocessing pipeline works?](docs/pipeline_preprocessing/pipeline.md)\n\n\n## Good pratices \n\n### PEP8\n\nUse the pep8 norm to format all the code. Specific pep8 parameters are given into the [pyproject.toml](pyproject.toml) file.\n\n```\nautopep8 --in-place --aggressive --aggressive ./\n```\n\n\n### FLAKE8\n\n[TO COMPLETE]\n',
    'author': 'Frederique Robin',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



from typing import Dict
import logging
from ..abstract_filter import SourceFilter


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


class SourceList(SourceFilter):
    """
        A basic source that generates a list.
    """

    def __init__(self, conf: Dict = {}):
        super().__init__()
        self.l_data = conf.get('l_data')

    def generator(self) -> object:
        logger.debug('Start generator')
        for elt in self.l_data:
            logger.debug(f'Element : {elt}')
            yield elt 
        logger.info('Stop generator')



from typing import Dict, List
from collections import OrderedDict
import copy

from .pipe import AbstractPipeline, Pipeline, SubPipeline, SUB_PIPELINE

from .filters.abstract_filter import SourceFilter, AbstractFilter
from . import GENERAL_CONF_PIPELINE

import logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


def construct_conf_subpipeline(conf: Dict, l_filters: List) -> Dict:
    conf_subpipeline = {}
    l_filters_type = []
    for filter_name in l_filters:
        conf_filter = conf.get(filter_name)
        conf_subpipeline[filter_name] = conf_filter
        l_filters_type.append(conf_filter.get('type'))
    return conf_subpipeline, l_filters_type


def create_filter(conf: Dict, catalog_filter: Dict, conf_global: Dict = None) -> AbstractFilter:
    _type = conf.get('type')
    _parms = conf.get('parms')
    try:
        if 'SubPipelineFilter' in _type:
            l_all_filters = copy.deepcopy(_parms.get('l_filters'))
            l_all_filters.append(conf.get('source'))
            conf_subpipeline, l_filters_type = construct_conf_subpipeline(conf_global, l_all_filters)
            _parms['conf_filter'] = conf_subpipeline
            _parms['catalog_filter'] = {filter_type: catalog_filter[filter_type] for filter_type in l_filters_type}
            _parms['conf_filter'][GENERAL_CONF_PIPELINE] = {'source': conf.get('source'),
                                                            'l_filters': _parms.get('l_filters'),
                                                            'sink': conf.get('sink')}
        return catalog_filter[_type](_parms)
    except KeyError as e:
        logger.info(f'Filter type {_type} does not exist. {e}')


def config_parser(conf: Dict) -> object:
    general_config = conf.get(GENERAL_CONF_PIPELINE)

    # Create source config dict
    source_name = general_config.get('source')
    source_config = conf.get(source_name)

    # Create sink config dict
    sink_name = general_config.get('sink')
    sink_config = conf.get(sink_name)

    # Create l_filters (ordered) config dict
    l_filters_name = general_config.get('l_filters')
    filters_config = OrderedDict()
    for filter_name in l_filters_name:
        filters_config[filter_name] = conf.get(filter_name)

    return source_config, filters_config, sink_config


def pipeline_factory(conf: Dict, catalog_filter: Dict = None, pipeline_type : str = None) -> AbstractPipeline:
    # Parse the conf file and create the Pipeline
    source_config, filters_config, sink_config = config_parser(conf)
    if pipeline_type == SUB_PIPELINE :
        pipeline = SubPipeline()
    else :
        pipeline = Pipeline()

    # Add source
    source = create_filter(source_config, catalog_filter)
    pipeline.add_source(source)

    # Add filters
    for _, conf_filter in filters_config.items():
        _filter = create_filter(conf_filter, catalog_filter, conf)
        pipeline.add_filter(_filter)

    # Add sink
    if sink_config is not None:
        sink = create_filter(sink_config, catalog_filter)
        pipeline.add_sink(sink)

    return pipeline
